.bulk-select-button-container {
    margin-top: 1rem;
}

.button-group {
    margin-bottom: .5rem;
}

.column-dropdown-checkbox {
    margin-right: .5rem;
}

.column-dropdown-label {
    width: 100%;
}

#column-selection-dropdown {
    float: right;
    margin-top: 1rem;
}

#experiment-table {
    width: 100%;
    padding: 0rem .25rem 1rem .25rem;
}

.experiment-table-header-text {
    padding-right: .5rem;
}

.external-link-icon {
    padding-left: .5rem;
}

.show-hide {
    display: none;
}
